@@ -7,6 +7,8 @@ package cache
 import (
 	"context"
 	"fmt"
+	"os"
+	"path/filepath"
 	"strconv"
 	"strings"
 	"sync"
@@ -25,7 +27,8 @@ type Session struct {
 	cache *Cache
 	id    string
 
-	options source.Options
+	optionsMu sync.Mutex
+	options   *source.Options
 
 	viewMu  sync.Mutex
 	views   []*View
@@ -46,7 +49,7 @@ type overlay struct {
 	version float64
 	kind    source.FileKind
 
-	// saved is true if a file has been saved on disk,
+	// saved is true if a file matches the state on disk,
 	// and therefore does not need to be part of the overlay sent to go/packages.
 	saved bool
 }
@@ -55,13 +58,19 @@ func (o *overlay) Read() ([]byte, error) {
 	return o.text, nil
 }
 
-func (o *overlay) Identity() source.FileIdentity {
+func (o *overlay) FileIdentity() source.FileIdentity {
 	return source.FileIdentity{
-		URI:        o.uri,
-		Identifier: o.hash,
-		SessionID:  o.session.id,
-		Version:    o.version,
-		Kind:       o.kind,
+		URI:  o.uri,
+		Hash: o.hash,
+		Kind: o.kind,
+	}
+}
+
+func (o *overlay) VersionedFileIdentity() source.VersionedFileIdentity {
+	return source.VersionedFileIdentity{
+		URI:       o.uri,
+		SessionID: o.session.id,
+		Version:   o.version,
 	}
 }
 
@@ -77,18 +86,47 @@ func (o *overlay) Version() float64 {
 	return o.version
 }
 
-func (o *overlay) Session() source.Session { return o.session }
-func (o *overlay) Saved() bool             { return o.saved }
-func (o *overlay) Data() []byte            { return o.text }
+func (o *overlay) Session() string {
+	return o.session.id
+}
+
+func (o *overlay) Saved() bool {
+	return o.saved
+}
+
+// closedFile implements LSPFile for a file that the editor hasn't told us about.
+type closedFile struct {
+	source.FileHandle
+}
+
+func (c *closedFile) VersionedFileIdentity() source.VersionedFileIdentity {
+	return source.VersionedFileIdentity{
+		URI:       c.FileHandle.URI(),
+		SessionID: "",
+		Version:   0,
+	}
+}
+
+func (c *closedFile) Session() string {
+	return ""
+}
+
+func (c *closedFile) Version() float64 {
+	return 0
+}
 
 func (s *Session) ID() string     { return s.id }
 func (s *Session) String() string { return s.id }
 
-func (s *Session) Options() source.Options {
+func (s *Session) Options() *source.Options {
+	s.optionsMu.Lock()
+	defer s.optionsMu.Unlock()
 	return s.options
 }
 
-func (s *Session) SetOptions(options source.Options) {
+func (s *Session) SetOptions(options *source.Options) {
+	s.optionsMu.Lock()
+	defer s.optionsMu.Unlock()
 	s.options = options
 }
 
@@ -103,79 +141,151 @@ func (s *Session) Shutdown(ctx context.Context) {
 	event.Log(ctx, "Shutdown session", KeyShutdownSession.Of(s))
 }
 
-func (s *Session) Cache() source.Cache {
+func (s *Session) Cache() interface{} {
 	return s.cache
 }
 
-func (s *Session) NewView(ctx context.Context, name string, folder span.URI, options source.Options) (source.View, source.Snapshot, error) {
+func (s *Session) NewView(ctx context.Context, name string, folder span.URI, options *source.Options) (source.View, source.Snapshot, func(), error) {
 	s.viewMu.Lock()
 	defer s.viewMu.Unlock()
-	v, snapshot, err := s.createView(ctx, name, folder, options, 0)
+	view, snapshot, release, err := s.createView(ctx, name, folder, options, 0)
 	if err != nil {
-		return nil, nil, err
+		return nil, nil, func() {}, err
 	}
-	s.views = append(s.views, v)
+	s.views = append(s.views, view)
 	// we always need to drop the view map
 	s.viewMap = make(map[span.URI]*View)
-	return v, snapshot, nil
+	return view, snapshot, release, nil
 }
 
-func (s *Session) createView(ctx context.Context, name string, folder span.URI, options source.Options, snapshotID uint64) (*View, *snapshot, error) {
+func (s *Session) createView(ctx context.Context, name string, folder span.URI, options *source.Options, snapshotID uint64) (*View, *snapshot, func(), error) {
 	index := atomic.AddInt64(&viewIndex, 1)
+
+	if s.cache.options != nil {
+		s.cache.options(options)
+	}
+
+	// Set the module-specific information.
+	ws, err := s.getWorkspaceInformation(ctx, folder, options)
+	if err != nil {
+		return nil, nil, func() {}, err
+	}
+
+	// If workspace module mode is enabled, find all of the modules in the
+	// workspace. By default, we just find the root module.
+	var modules map[span.URI]*moduleRoot
+	modules, err = findWorkspaceModules(ctx, ws.rootURI, options)
+	if err != nil {
+		return nil, nil, func() {}, err
+	}
+
 	// We want a true background context and not a detached context here
 	// the spans need to be unrelated and no tag values should pollute it.
 	baseCtx := event.Detach(xcontext.Detach(ctx))
 	backgroundCtx, cancel := context.WithCancel(baseCtx)
 
 	v := &View{
-		session:            s,
-		initialized:        make(chan struct{}),
-		initializationSema: make(chan struct{}, 1),
-		initializeOnce:     &sync.Once{},
-		id:                 strconv.FormatInt(index, 10),
-		options:            options,
-		baseCtx:            baseCtx,
-		backgroundCtx:      backgroundCtx,
-		cancel:             cancel,
-		name:               name,
-		folder:             folder,
-		filesByURI:         make(map[span.URI]*fileBase),
-		filesByBase:        make(map[string][]*fileBase),
-		snapshot: &snapshot{
-			id:                snapshotID,
-			packages:          make(map[packageKey]*packageHandle),
-			ids:               make(map[span.URI][]packageID),
-			metadata:          make(map[packageID]*metadata),
-			files:             make(map[span.URI]source.FileHandle),
-			importedBy:        make(map[packageID][]packageID),
-			actions:           make(map[actionKey]*actionHandle),
-			workspacePackages: make(map[packageID]packagePath),
-			unloadableFiles:   make(map[span.URI]struct{}),
-			parseModHandles:   make(map[span.URI]*parseModHandle),
+		session:              s,
+		initialized:          make(chan struct{}),
+		initializationSema:   make(chan struct{}, 1),
+		initializeOnce:       &sync.Once{},
+		id:                   strconv.FormatInt(index, 10),
+		options:              options,
+		baseCtx:              baseCtx,
+		backgroundCtx:        backgroundCtx,
+		cancel:               cancel,
+		name:                 name,
+		folder:               folder,
+		filesByURI:           make(map[span.URI]*fileBase),
+		filesByBase:          make(map[string][]*fileBase),
+		workspaceInformation: *ws,
+	}
+	v.importsState = &importsState{
+		ctx: backgroundCtx,
+		processEnv: &imports.ProcessEnv{
+			GocmdRunner: s.gocmdRunner,
+			WorkingDir:  folder.Filename(),
+			Env:         ws.goEnv,
 		},
 	}
-	v.snapshot.view = v
-
-	if v.session.cache.options != nil {
-		v.session.cache.options(&v.options)
-	}
-	// Set the module-specific information.
-	if err := v.setBuildInformation(ctx, folder, options.Env, v.options.TempModfile); err != nil {
-		return nil, nil, err
+	v.snapshot = &snapshot{
+		id:                snapshotID,
+		view:              v,
+		generation:        s.cache.store.Generation(generationName(v, 0)),
+		packages:          make(map[packageKey]*packageHandle),
+		ids:               make(map[span.URI][]packageID),
+		metadata:          make(map[packageID]*metadata),
+		files:             make(map[span.URI]source.VersionedFileHandle),
+		goFiles:           make(map[parseKey]*parseGoHandle),
+		importedBy:        make(map[packageID][]packageID),
+		actions:           make(map[actionKey]*actionHandle),
+		workspacePackages: make(map[packageID]packagePath),
+		unloadableFiles:   make(map[span.URI]struct{}),
+		parseModHandles:   make(map[span.URI]*parseModHandle),
+		modTidyHandles:    make(map[span.URI]*modTidyHandle),
+		modUpgradeHandles: make(map[span.URI]*modUpgradeHandle),
+		modWhyHandles:     make(map[span.URI]*modWhyHandle),
+		modules:           modules,
 	}
 
-	// We have v.goEnv now.
-	v.processEnv = &imports.ProcessEnv{
-		GocmdRunner: s.gocmdRunner,
-		WorkingDir:  folder.Filename(),
-		Env:         v.goEnv,
-	}
+	v.snapshot.workspaceDirectories = v.snapshot.findWorkspaceDirectories(ctx)
 
 	// Initialize the view without blocking.
 	initCtx, initCancel := context.WithCancel(xcontext.Detach(ctx))
 	v.initCancelFirstAttempt = initCancel
-	go v.initialize(initCtx, v.snapshot, true)
-	return v, v.snapshot, nil
+	snapshot := v.snapshot
+	release := snapshot.generation.Acquire(initCtx)
+	go func() {
+		snapshot.initialize(initCtx, true)
+		release()
+	}()
+	return v, snapshot, snapshot.generation.Acquire(ctx), nil
+}
+
+// findWorkspaceModules walks the view's root folder, looking for go.mod files.
+// Any that are found are added to the view's set of modules, which are then
+// used to construct the workspace module.
+//
+// It assumes that the caller has not yet created the view, and therefore does
+// not lock any of the internal data structures before accessing them.
+//
+// TODO(rstambler): Check overlays for go.mod files.
+func findWorkspaceModules(ctx context.Context, root span.URI, options *source.Options) (map[span.URI]*moduleRoot, error) {
+	// Walk the view's folder to find all modules in the view.
+	modules := make(map[span.URI]*moduleRoot)
+	if !options.ExperimentalWorkspaceModule {
+		path := filepath.Join(root.Filename(), "go.mod")
+		if info, _ := os.Stat(path); info != nil {
+			if m := getViewModule(ctx, root, span.URIFromPath(path), options); m != nil {
+				modules[m.rootURI] = m
+			}
+		}
+		return modules, nil
+	}
+	return modules, filepath.Walk(root.Filename(), func(path string, info os.FileInfo, err error) error {
+		if err != nil {
+			// Probably a permission error. Keep looking.
+			return filepath.SkipDir
+		}
+		// For any path that is not the workspace folder, check if the path
+		// would be ignored by the go command. Vendor directories also do not
+		// contain workspace modules.
+		if info.IsDir() && path != root.Filename() {
+			suffix := strings.TrimPrefix(path, root.Filename())
+			switch {
+			case checkIgnored(suffix),
+				strings.Contains(filepath.ToSlash(suffix), "/vendor/"):
+				return filepath.SkipDir
+			}
+		}
+		// We're only interested in go.mod files.
+		if filepath.Base(path) == "go.mod" {
+			if m := getViewModule(ctx, root, span.URIFromPath(path), options); m != nil {
+				modules[m.rootURI] = m
+			}
+		}
+		return nil
+	})
 }
 
 // View returns the view by name.
@@ -280,18 +390,19 @@ func (s *Session) removeView(ctx context.Context, view *View) error {
 	return nil
 }
 
-func (s *Session) updateView(ctx context.Context, view *View, options source.Options) (*View, *snapshot, error) {
+func (s *Session) updateView(ctx context.Context, view *View, options *source.Options) (*View, error) {
 	s.viewMu.Lock()
 	defer s.viewMu.Unlock()
 	i, err := s.dropView(ctx, view)
 	if err != nil {
-		return nil, nil, err
+		return nil, err
 	}
 	// Preserve the snapshot ID if we are recreating the view.
 	view.snapshotMu.Lock()
 	snapshotID := view.snapshot.id
 	view.snapshotMu.Unlock()
-	v, snapshot, err := s.createView(ctx, view.name, view.folder, options, snapshotID)
+	v, _, release, err := s.createView(ctx, view.name, view.folder, options, snapshotID)
+	release()
 	if err != nil {
 		// we have dropped the old view, but could not create the new one
 		// this should not happen and is very bad, but we still need to clean
@@ -299,10 +410,11 @@ func (s *Session) updateView(ctx context.Context, view *View, options source.Opt
 		s.views[i] = s.views[len(s.views)-1]
 		s.views[len(s.views)-1] = nil
 		s.views = s.views[:len(s.views)-1]
+		return nil, err
 	}
 	// substitute the new view into the array where the old view was
 	s.views[i] = v
-	return v, snapshot, nil
+	return v, nil
 }
 
 func (s *Session) dropView(ctx context.Context, v *View) (int, error) {
@@ -319,69 +431,95 @@ func (s *Session) dropView(ctx context.Context, v *View) (int, error) {
 	return -1, errors.Errorf("view %s for %v not found", v.Name(), v.Folder())
 }
 
-func (s *Session) DidModifyFiles(ctx context.Context, changes []source.FileModification) ([]source.Snapshot, error) {
-	views := make(map[*View]map[span.URI]source.FileHandle)
+func (s *Session) ModifyFiles(ctx context.Context, changes []source.FileModification) error {
+	_, _, releases, _, err := s.DidModifyFiles(ctx, changes)
+	for _, release := range releases {
+		release()
+	}
+	return err
+}
+
+func (s *Session) DidModifyFiles(ctx context.Context, changes []source.FileModification) (map[span.URI]source.View, map[source.View]source.Snapshot, []func(), []span.URI, error) {
+	views := make(map[*View]map[span.URI]source.VersionedFileHandle)
+	bestViews := map[span.URI]source.View{}
+	// Keep track of deleted files so that we can clear their diagnostics.
+	// A file might be re-created after deletion, so only mark files that
+	// have truly been deleted.
+	deletions := map[span.URI]struct{}{}
 
 	overlays, err := s.updateOverlays(ctx, changes)
 	if err != nil {
-		return nil, err
+		return nil, nil, nil, nil, err
 	}
 	var forceReloadMetadata bool
 	for _, c := range changes {
 		if c.Action == source.InvalidateMetadata {
 			forceReloadMetadata = true
 		}
-		// Look through all of the session's views, invalidating the file for
-		// all of the views to which it is known.
+
+		// Build the list of affected views.
+		bestView, err := s.viewOf(c.URI)
+		if err != nil {
+			return nil, nil, nil, nil, err
+		}
+		bestViews[c.URI] = bestView
+
+		var changedViews []*View
 		for _, view := range s.views {
 			// Don't propagate changes that are outside of the view's scope
 			// or knowledge.
 			if !view.relevantChange(c) {
 				continue
 			}
+			changedViews = append(changedViews, view)
+		}
+		// If no view matched the change, assign it to the best view.
+		if len(changedViews) == 0 {
+			changedViews = append(changedViews, bestView)
+		}
+
+		// Apply the changes to all affected views.
+		for _, view := range changedViews {
 			// Make sure that the file is added to the view.
 			if _, err := view.getFile(c.URI); err != nil {
-				return nil, err
+				return nil, nil, nil, nil, err
 			}
 			if _, ok := views[view]; !ok {
-				views[view] = make(map[span.URI]source.FileHandle)
+				views[view] = make(map[span.URI]source.VersionedFileHandle)
 			}
 			var (
-				fh  source.FileHandle
-				ok  bool
-				err error
+				fh source.VersionedFileHandle
+				ok bool
 			)
 			if fh, ok = overlays[c.URI]; ok {
 				views[view][c.URI] = fh
+				delete(deletions, c.URI)
 			} else {
-				fh, err = s.cache.getFile(ctx, c.URI)
+				fsFile, err := s.cache.getFile(ctx, c.URI)
 				if err != nil {
-					return nil, err
+					return nil, nil, nil, nil, err
 				}
+				fh = &closedFile{fsFile}
 				views[view][c.URI] = fh
-			}
-			// If the file change is to a go.mod file, and initialization for
-			// the view has previously failed, we should attempt to retry.
-			// TODO(rstambler): We can use unsaved contents with -modfile, so
-			// maybe we should do that and retry on any change?
-			if fh.Kind() == source.Mod && (c.OnDisk || c.Action == source.Save) {
-				view.maybeReinitialize()
+				if _, err := fh.Read(); err != nil {
+					deletions[c.URI] = struct{}{}
+				}
 			}
 		}
 	}
-	var snapshots []source.Snapshot
+
+	snapshots := map[source.View]source.Snapshot{}
+	var releases []func()
 	for view, uris := range views {
-		snapshots = append(snapshots, view.invalidateContent(ctx, uris, forceReloadMetadata))
+		snapshot, release := view.invalidateContent(ctx, uris, forceReloadMetadata)
+		snapshots[view] = snapshot
+		releases = append(releases, release)
 	}
-	return snapshots, nil
-}
-
-func (s *Session) isOpen(uri span.URI) bool {
-	s.overlayMu.Lock()
-	defer s.overlayMu.Unlock()
-
-	_, open := s.overlays[uri]
-	return open
+	var deletionsSlice []span.URI
+	for uri := range deletions {
+		deletionsSlice = append(deletionsSlice, uri)
+	}
+	return bestViews, snapshots, releases, deletionsSlice, nil
 }
 
 func (s *Session) updateOverlays(ctx context.Context, changes []source.FileModification) (map[span.URI]*overlay, error) {
@@ -459,7 +597,7 @@ func (s *Session) updateOverlays(ctx context.Context, changes []source.FileModif
 				return nil, err
 			}
 			_, readErr := fh.Read()
-			sameContentOnDisk = (readErr == nil && fh.Identity().Identifier == hash)
+			sameContentOnDisk = (readErr == nil && fh.FileIdentity().Hash == hash)
 		}
 		o = &overlay{
 			session: s,
