@@ -130,8 +130,10 @@ outer:
 		}
 
 		// If obj is not accessible because it lives in another package and is
-		// not exported, don't treat it as a completion candidate.
-		if obj.Pkg() != nil && obj.Pkg() != c.pkg.GetTypes() && !obj.Exported() {
+		// not exported, don't treat it as a completion candidate unless it's
+		// a package completion candidate.
+		if !c.completionContext.packageCompletion &&
+			obj.Pkg() != nil && obj.Pkg() != c.pkg.GetTypes() && !obj.Exported() {
 			continue
 		}
 
@@ -196,7 +198,7 @@ outer:
 			if sig.Params().Len() == 0 && sig.Results().Len() == 1 {
 				path, names := c.deepState.newPath(cand, obj, true)
 				// The result of a function call is not addressable.
-				candidates := c.methodsAndFields(ctx, sig.Results().At(0).Type(), false, cand.imp)
+				candidates := c.methodsAndFields(sig.Results().At(0).Type(), false, cand.imp)
 				for _, newCand := range candidates {
 					newCand.path, newCand.names = path, names
 					c.deepState.enqueue(newCand)
@@ -207,13 +209,13 @@ outer:
 		path, names := c.deepState.newPath(cand, obj, false)
 		switch obj := obj.(type) {
 		case *types.PkgName:
-			candidates := c.packageMembers(ctx, obj.Imported(), stdScore, cand.imp)
+			candidates := c.packageMembers(obj.Imported(), stdScore, cand.imp)
 			for _, newCand := range candidates {
 				newCand.path, newCand.names = path, names
 				c.deepState.enqueue(newCand)
 			}
 		default:
-			candidates := c.methodsAndFields(ctx, obj.Type(), cand.addressable, cand.imp)
+			candidates := c.methodsAndFields(obj.Type(), cand.addressable, cand.imp)
 			for _, newCand := range candidates {
 				newCand.path, newCand.names = path, names
 				c.deepState.enqueue(newCand)
@@ -262,19 +264,9 @@ func (c *completer) addCandidate(ctx context.Context, cand *candidate) {
 	}
 
 	cand.name = strings.Join(append(cand.names, cand.obj.Name()), ".")
-	matchScore := c.matcher.Score(cand.name)
-	if matchScore > 0 {
-		cand.score *= float64(matchScore)
-
-		// Avoid calling c.item() for deep candidates that wouldn't be in the top
-		// MaxDeepCompletions anyway.
-		if len(cand.path) == 0 || c.deepState.isHighScore(cand.score) {
-			if item, err := c.item(ctx, *cand); err == nil {
-				c.items = append(c.items, item)
-			}
-		}
+	if item, err := c.item(ctx, *cand); err == nil {
+		c.items = append(c.items, item)
 	}
-
 }
 
 // penalty reports a score penalty for cand in the range (0, 1).
