@@ -59,7 +59,7 @@ var (
 	defaultOptions *Options
 )
 
-//go:generate go run golang.org/x/tools/internal/lsp/source/genopts -output options_json.go
+//go:generate go run golang.org/x/tools/internal/lsp/source/genapijson -output api_json.go
 
 // DefaultOptions is the options that are used for Gopls execution independent
 // of any externally provided configuration (LSP initialization, command
@@ -68,7 +68,7 @@ func DefaultOptions() *Options {
 	optionsOnce.Do(func() {
 		var commands []string
 		for _, c := range Commands {
-			commands = append(commands, c.Name)
+			commands = append(commands, c.ID())
 		}
 		defaultOptions = &Options{
 			ClientOptions: ClientOptions{
@@ -101,8 +101,7 @@ func DefaultOptions() *Options {
 				LinkTarget: "pkg.go.dev",
 			},
 			DebuggingOptions: DebuggingOptions{
-				CompletionBudget:   100 * time.Millisecond,
-				LiteralCompletions: true,
+				CompletionBudget: 100 * time.Millisecond,
 			},
 			ExperimentalOptions: ExperimentalOptions{
 				TempModfile:             true,
@@ -124,6 +123,9 @@ func DefaultOptions() *Options {
 				SymbolMatcher:           SymbolFuzzy,
 				SymbolStyle:             PackageQualifiedSymbols,
 			},
+			InternalOptions: InternalOptions{
+				LiteralCompletions: true,
+			},
 			Hooks: Hooks{
 				ComputeEdits:         myers.ComputeEdits,
 				URLRegexp:            urlRegexp(),
@@ -146,6 +148,7 @@ type Options struct {
 	UserOptions
 	DebuggingOptions
 	ExperimentalOptions
+	InternalOptions
 	Hooks
 }
 
@@ -177,7 +180,7 @@ type UserOptions struct {
 	BuildFlags []string
 
 	// Env adds environment variables to external commands run by `gopls`, most notably `go list`.
-	Env []string
+	Env map[string]string
 
 	// HoverKind controls the information that appears in the hover text.
 	// SingleLine and Structured are intended for use only by authors of editor plugins.
@@ -203,6 +206,27 @@ type UserOptions struct {
 	Gofumpt bool
 }
 
+// EnvSlice returns Env as a slice of k=v strings.
+func (u *UserOptions) EnvSlice() []string {
+	var result []string
+	for k, v := range u.Env {
+		result = append(result, fmt.Sprintf("%v=%v", k, v))
+	}
+	return result
+}
+
+// SetEnvSlice sets Env from a slice of k=v strings.
+func (u *UserOptions) SetEnvSlice(env []string) {
+	u.Env = map[string]string{}
+	for _, kv := range env {
+		split := strings.SplitN(kv, "=", 2)
+		if len(split) != 2 {
+			continue
+		}
+		u.Env[split[0]] = split[1]
+	}
+}
+
 // Hooks contains configuration that is provided to the Gopls command by the
 // main package.
 type Hooks struct {
@@ -235,20 +259,15 @@ type ExperimentalOptions struct {
 	// ```
 	Analyses map[string]bool
 
-	// Overrides the enabled/disabled state of various code lenses. Currently, we
-	// support several code lenses:
-	//
-	// * `generate`: run `go generate` as specified by a `//go:generate` directive.
-	// * `upgrade_dependency`: upgrade a dependency listed in a `go.mod` file.
-	// * `test`: run `go test -run` for a test func.
-	// * `gc_details`: Show the gc compiler's choices for inline analysis and escaping.
+	// Codelens overrides the enabled/disabled state of code lenses. See the "Code Lenses"
+	// section of settings.md for the list of supported lenses.
 	//
 	// Example Usage:
 	// ```json5
 	// "gopls": {
 	// ...
 	//   "codelens": {
-	//     "generate": false,  // Don't run `go generate`.
+	//     "generate": false,  // Don't show the `go generate` lens.
 	//     "gc_details": true  // Show a code lens toggling the display of gc's choices.
 	//   }
 	// ...
@@ -262,7 +281,7 @@ type ExperimentalOptions struct {
 	// CompleteUnimported enables completion for packages that you do not currently import.
 	CompleteUnimported bool
 
-	// DeepCompletion If true, this turns on the ability to return completions from deep inside relevant entities, rather than just the locally accessible ones.
+	// DeepCompletion enables the ability to return completions from deep inside relevant entities, rather than just the locally accessible ones.
 	//
 	// Consider this example:
 	//
@@ -301,7 +320,16 @@ type ExperimentalOptions struct {
 	// SymbolMatcher sets the algorithm that is used when finding workspace symbols.
 	SymbolMatcher SymbolMatcher
 
-	// SymbolStyle specifies what style of symbols to return in symbol requests.
+	// SymbolStyle controls how symbols are qualified in symbol responses.
+	//
+	// Example Usage:
+	// ```json5
+	// "gopls": {
+	// ...
+	//   "symbolStyle": "dynamic",
+	// ...
+	// }
+	// ```
 	SymbolStyle SymbolStyle
 
 	// LinksInHover toggles the presence of links to documentation in hover.
@@ -318,6 +346,10 @@ type ExperimentalOptions struct {
 	// progress reports for all work done outside the scope of an RPC.
 	VerboseWorkDoneProgress bool
 
+	// SemanticTokens controls whether the LSP server will send
+	// semantic tokens to the client.
+	SemanticTokens bool
+
 	// ExpandWorkspaceToModule instructs `gopls` to expand the scope of the workspace to include the
 	// modules containing the workspace folders. Set this to false to avoid loading
 	// your entire module. This is particularly useful for those working in a monorepo.
@@ -326,6 +358,23 @@ type ExperimentalOptions struct {
 	// ExperimentalWorkspaceModule opts a user into the experimental support
 	// for multi-module workspaces.
 	ExperimentalWorkspaceModule bool
+
+	// ExperimentalDiagnosticsDelay controls the amount of time that gopls waits
+	// after the most recent file modification before computing deep diagnostics.
+	// Simple diagnostics (parsing and type-checking) are always run immediately
+	// on recently modified packages.
+	//
+	// This option must be set to a valid duration string, for example `"250ms"`.
+	ExperimentalDiagnosticsDelay time.Duration
+
+	// ExperimentalPackageCacheKey controls whether to use a coarser cache key
+	// for package type information to increase cache hits. This setting removes
+	// the user's environment, build flags, and working directory from the cache
+	// key, which should be a safe change as all relevant inputs into the type
+	// checking pass are already hashed into the key. This is temporarily guarded
+	// by an experiment because caching behavior is subtle and difficult to
+	// comprehensively test.
+	ExperimentalPackageCacheKey bool
 }
 
 // DebuggingOptions should not affect the logical execution of Gopls, but may
@@ -340,7 +389,11 @@ type DebuggingOptions struct {
 	// dynamically reduce the search scope to ensure we return timely
 	// results. Zero means unlimited.
 	CompletionBudget time.Duration
+}
 
+// InternalOptions contains settings that are not exposed to the user for various
+// reasons, e.g. settings used by tests.
+type InternalOptions struct {
 	// LiteralCompletions controls whether literal candidates such as
 	// "&someStruct{}" are offered. Tests disable this flag to simplify
 	// their expected values.
@@ -382,9 +435,17 @@ const (
 type SymbolStyle string
 
 const (
+	// PackageQualifiedSymbols is package qualified symbols i.e.
+	// "pkg.Foo.Field".
 	PackageQualifiedSymbols SymbolStyle = "Package"
-	FullyQualifiedSymbols   SymbolStyle = "Full"
-	DynamicSymbols          SymbolStyle = "Dynamic"
+	// FullyQualifiedSymbols is fully qualified symbols, i.e.
+	// "path/to/pkg.Foo.Field".
+	FullyQualifiedSymbols SymbolStyle = "Full"
+	// DynamicSymbols uses whichever qualifier results in the highest scoring
+	// match for the given symbol query. Here a "qualifier" is any "/" or "."
+	// delimited suffix of the fully qualified symbol. i.e. "to/pkg.Foo.Field" or
+	// just "Foo.Field".
+	DynamicSymbols SymbolStyle = "Dynamic"
 )
 
 type HoverKind string
@@ -429,9 +490,23 @@ func SetOptions(options *Options, opts interface{}) OptionResults {
 	switch opts := opts.(type) {
 	case nil:
 	case map[string]interface{}:
+		// If the user's settings contains "allExperiments", set that first,
+		// and then let them override individual settings independently.
+		var enableExperiments bool
+		for name, value := range opts {
+			if b, ok := value.(bool); name == "allExperiments" && ok && b {
+				enableExperiments = true
+				options.enableAllExperiments()
+			}
+		}
 		for name, value := range opts {
 			results = append(results, options.set(name, value))
 		}
+		// Finally, enable any experimental features that are specified in
+		// maps, which allows users to individually toggle them on or off.
+		if enableExperiments {
+			options.enableAllExperimentMaps()
+		}
 	default:
 		results = append(results, OptionResult{
 			Value: opts,
@@ -467,6 +542,7 @@ func (o *Options) Clone() *Options {
 		ClientOptions:       o.ClientOptions,
 		DebuggingOptions:    o.DebuggingOptions,
 		ExperimentalOptions: o.ExperimentalOptions,
+		InternalOptions:     o.InternalOptions,
 		Hooks: Hooks{
 			GoDiff:        o.Hooks.GoDiff,
 			ComputeEdits:  o.Hooks.ComputeEdits,
@@ -494,7 +570,7 @@ func (o *Options) Clone() *Options {
 		copy(dst, src)
 		return dst
 	}
-	result.Env = copySlice(o.Env)
+	result.SetEnvSlice(o.EnvSlice())
 	result.BuildFlags = copySlice(o.BuildFlags)
 
 	copyAnalyzerMap := func(src map[string]Analyzer) map[string]Analyzer {
@@ -511,8 +587,27 @@ func (o *Options) Clone() *Options {
 	return result
 }
 
-func (options *Options) AddStaticcheckAnalyzer(a *analysis.Analyzer) {
-	options.StaticcheckAnalyzers[a.Name] = Analyzer{Analyzer: a, Enabled: true}
+func (o *Options) AddStaticcheckAnalyzer(a *analysis.Analyzer) {
+	o.StaticcheckAnalyzers[a.Name] = Analyzer{Analyzer: a, Enabled: true}
+}
+
+// enableAllExperiments turns on all of the experimental "off-by-default"
+// features offered by gopls.
+// Any experimental features specified in maps should be enabled in
+// enableAllExperimentMaps.
+func (o *Options) enableAllExperiments() {
+	o.ExperimentalDiagnosticsDelay = 200 * time.Millisecond
+	o.ExperimentalPackageCacheKey = true
+	o.SymbolStyle = DynamicSymbols
+}
+
+func (o *Options) enableAllExperimentMaps() {
+	if _, ok := o.Codelens[CommandToggleDetails.Name]; !ok {
+		o.Codelens[CommandToggleDetails.Name] = true
+	}
+	if _, ok := o.Analyses[unusedparams.Analyzer.Name]; !ok {
+		o.Analyses[unusedparams.Analyzer.Name] = true
+	}
 }
 
 func (o *Options) set(name string, value interface{}) OptionResult {
@@ -521,17 +616,17 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 	case "env":
 		menv, ok := value.(map[string]interface{})
 		if !ok {
-			result.errorf("invalid config gopls.env type %T", value)
+			result.errorf("invalid type %T, expect map", value)
 			break
 		}
 		for k, v := range menv {
-			o.Env = append(o.Env, fmt.Sprintf("%s=%s", k, v))
+			o.Env[k] = fmt.Sprint(v)
 		}
 
 	case "buildFlags":
 		iflags, ok := value.([]interface{})
 		if !ok {
-			result.errorf("invalid config gopls.buildFlags type %T", value)
+			result.errorf("invalid type %T, expect list", value)
 			break
 		}
 		flags := make([]string, 0, len(iflags))
@@ -549,77 +644,43 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 	case "completeUnimported":
 		result.setBool(&o.CompleteUnimported)
 	case "completionBudget":
-		if v, ok := result.asString(); ok {
-			d, err := time.ParseDuration(v)
-			if err != nil {
-				result.errorf("failed to parse duration %q: %v", v, err)
-				break
-			}
-			o.CompletionBudget = d
-		}
-
+		result.setDuration(&o.CompletionBudget)
 	case "matcher":
-		matcher, ok := result.asString()
-		if !ok {
-			break
-		}
-		switch strings.ToLower(matcher) {
-		case "fuzzy":
-			o.Matcher = Fuzzy
-		case "casesensitive":
-			o.Matcher = CaseSensitive
-		default:
-			o.Matcher = CaseInsensitive
+		if s, ok := result.asOneOf(
+			string(Fuzzy),
+			string(CaseSensitive),
+			string(CaseInsensitive),
+		); ok {
+			o.Matcher = Matcher(s)
 		}
 
 	case "symbolMatcher":
-		matcher, ok := result.asString()
-		if !ok {
-			break
-		}
-		switch strings.ToLower(matcher) {
-		case "fuzzy":
-			o.SymbolMatcher = SymbolFuzzy
-		case "casesensitive":
-			o.SymbolMatcher = SymbolCaseSensitive
-		default:
-			o.SymbolMatcher = SymbolCaseInsensitive
+		if s, ok := result.asOneOf(
+			string(SymbolFuzzy),
+			string(SymbolCaseInsensitive),
+			string(SymbolCaseSensitive),
+		); ok {
+			o.SymbolMatcher = SymbolMatcher(s)
 		}
 
 	case "symbolStyle":
-		style, ok := result.asString()
-		if !ok {
-			break
-		}
-		switch strings.ToLower(style) {
-		case "full":
-			o.SymbolStyle = FullyQualifiedSymbols
-		case "dynamic":
-			o.SymbolStyle = DynamicSymbols
-		case "package":
-			o.SymbolStyle = PackageQualifiedSymbols
-		default:
-			result.errorf("Unsupported symbol style %q", style)
+		if s, ok := result.asOneOf(
+			string(FullyQualifiedSymbols),
+			string(PackageQualifiedSymbols),
+			string(DynamicSymbols),
+		); ok {
+			o.SymbolStyle = SymbolStyle(s)
 		}
 
 	case "hoverKind":
-		hoverKind, ok := result.asString()
-		if !ok {
-			break
-		}
-		switch strings.ToLower(hoverKind) {
-		case "nodocumentation":
-			o.HoverKind = NoDocumentation
-		case "singleline":
-			o.HoverKind = SingleLine
-		case "synopsisdocumentation":
-			o.HoverKind = SynopsisDocumentation
-		case "fulldocumentation":
-			o.HoverKind = FullDocumentation
-		case "structured":
-			o.HoverKind = Structured
-		default:
-			result.errorf("Unsupported hover kind %q", hoverKind)
+		if s, ok := result.asOneOf(
+			string(NoDocumentation),
+			string(SingleLine),
+			string(SynopsisDocumentation),
+			string(FullDocumentation),
+			string(Structured),
+		); ok {
+			o.HoverKind = HoverKind(s)
 		}
 
 	case "linkTarget":
@@ -629,15 +690,8 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 		result.setBool(&o.LinksInHover)
 
 	case "importShortcut":
-		var s string
-		result.setString(&s)
-		switch strings.ToLower(s) {
-		case "both":
-			o.ImportShortcut = Both
-		case "link":
-			o.ImportShortcut = Link
-		case "definition":
-			o.ImportShortcut = Definition
+		if s, ok := result.asOneOf(string(Both), string(Link), string(Definition)); ok {
+			o.ImportShortcut = ImportShortcut(s)
 		}
 
 	case "analyses":
@@ -685,12 +739,25 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 	case "gofumpt":
 		result.setBool(&o.Gofumpt)
 
+	case "semanticTokens":
+		result.setBool(&o.SemanticTokens)
+
 	case "expandWorkspaceToModule":
 		result.setBool(&o.ExpandWorkspaceToModule)
 
 	case "experimentalWorkspaceModule":
 		result.setBool(&o.ExperimentalWorkspaceModule)
 
+	case "experimentalDiagnosticsDelay":
+		result.setDuration(&o.ExperimentalDiagnosticsDelay)
+
+	case "experimentalPackageCacheKey":
+		result.setBool(&o.ExperimentalPackageCacheKey)
+
+	case "allExperiments":
+		// This setting should be handled before all of the other options are
+		// processed, so do nothing here.
+
 	// Replaced settings.
 	case "experimentalDisabledAnalyses":
 		result.State = OptionDeprecated
@@ -740,13 +807,14 @@ func (o *Options) set(name string, value interface{}) OptionResult {
 }
 
 func (r *OptionResult) errorf(msg string, values ...interface{}) {
-	r.Error = errors.Errorf(msg, values...)
+	prefix := fmt.Sprintf("parsing setting %q: ", r.Name)
+	r.Error = errors.Errorf(prefix+msg, values...)
 }
 
 func (r *OptionResult) asBool() (bool, bool) {
 	b, ok := r.Value.(bool)
 	if !ok {
-		r.errorf("Invalid type %T for bool option %q", r.Value, r.Name)
+		r.errorf("invalid type %T, expect bool", r.Value)
 		return false, false
 	}
 	return b, true
@@ -758,10 +826,21 @@ func (r *OptionResult) setBool(b *bool) {
 	}
 }
 
+func (r *OptionResult) setDuration(d *time.Duration) {
+	if v, ok := r.asString(); ok {
+		parsed, err := time.ParseDuration(v)
+		if err != nil {
+			r.errorf("failed to parse duration %q: %v", v, err)
+			return
+		}
+		*d = parsed
+	}
+}
+
 func (r *OptionResult) setBoolMap(bm *map[string]bool) {
 	all, ok := r.Value.(map[string]interface{})
 	if !ok {
-		r.errorf("Invalid type %T for map[string]interface{} option %q", r.Value, r.Name)
+		r.errorf("invalid type %T for map[string]bool option", r.Value)
 		return
 	}
 	m := make(map[string]bool)
@@ -769,7 +848,7 @@ func (r *OptionResult) setBoolMap(bm *map[string]bool) {
 		if enabled, ok := enabled.(bool); ok {
 			m[a] = enabled
 		} else {
-			r.errorf("Invalid type %d for map key %q in option %q", a, r.Name)
+			r.errorf("invalid type %T for map key %q", enabled, a)
 			return
 		}
 	}
@@ -779,12 +858,27 @@ func (r *OptionResult) setBoolMap(bm *map[string]bool) {
 func (r *OptionResult) asString() (string, bool) {
 	b, ok := r.Value.(string)
 	if !ok {
-		r.errorf("Invalid type %T for string option %q", r.Value, r.Name)
+		r.errorf("invalid type %T, expect string", r.Value)
 		return "", false
 	}
 	return b, true
 }
 
+func (r *OptionResult) asOneOf(options ...string) (string, bool) {
+	s, ok := r.asString()
+	if !ok {
+		return "", false
+	}
+	lower := strings.ToLower(s)
+	for _, opt := range options {
+		if strings.ToLower(opt) == lower {
+			return opt, true
+		}
+	}
+	r.errorf("invalid option %q for enum", r.Value)
+	return "", false
+}
+
 func (r *OptionResult) setString(s *string) {
 	if v, ok := r.asString(); ok {
 		*s = v
@@ -900,3 +994,34 @@ func urlRegexp() *regexp.Regexp {
 	re.Longest()
 	return re
 }
+
+type APIJSON struct {
+	Options  map[string][]*OptionJSON
+	Commands []*CommandJSON
+	Lenses   []*LensJSON
+}
+
+type OptionJSON struct {
+	Name       string
+	Type       string
+	Doc        string
+	EnumValues []EnumValue
+	Default    string
+}
+
+type EnumValue struct {
+	Value string
+	Doc   string
+}
+
+type CommandJSON struct {
+	Command string
+	Title   string
+	Doc     string
+}
+
+type LensJSON struct {
+	Lens  string
+	Title string
+	Doc   string
+}
